from __future__ import annotations

import json
from pathlib import Path

from aws_cdk import aws_apigateway as apigateway


_cache_dir = Path('~/.cdk/cache').expanduser()
_cache_dir.mkdir(parents=True, exist_ok=True)

_cache = _cache_dir / 'apigw_api_keys.json'

try:
    with _cache.open() as in_file:
        _keys = json.load(in_file)

except FileNotFoundError:
    _keys = {}


class APIKeyCache:

    __slots__ = (
        '_account',
        '_stack_name',
    )

    def __init__(self, construct: apigateway.IRestApi):
        self._account = construct.env.account
        self._stack_name = construct.stack.stack_name

    def get_api_key(self) -> str | None:
        """Retrieve the cached API key value for the stack's construct."""
        try:
            return _keys[self._account][self._stack_name]
        except KeyError:
            return None

    def save_api_key(self, api_key: str):
        """Store (cache) the API key value for the stack's construct."""
        _keys.setdefault(self._account, {})[self._stack_name] = api_key

        with _cache.open('w') as out_file:
            json.dump(_keys, out_file)



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum, auto

from aws_cdk import aws_apigateway as apigateway


@dataclass
class Config:
    # AWS Region Name, ex. `us-east-1`
    region: str | None = None

    # AWS Profile Name, ex. `my-aws-profile`
    profile: str | None = None

    # `key_id` - The Key Management Service (KMS) ID that you want to
    # use to encrypt a parameter. Either the default KMS key automatically
    # assigned to your AWS account or a custom key. Required for parameters
    # that use the `SecureString` data type.
    #
    # If you don't specify a key ID, the system uses the default key
    # associated with your AWS account.
    key_id: str | None = None

    # Whether to override the CDK stack output name for the API endpoint.
    # Defaults to `True` if not specified.
    override_endpoint_name = True

    # Throttle settings to associate with the Usage Plan for the REST API
    throttle: apigateway.ThrottleSettings = apigateway.ThrottleSettings(
        burst_limit=500,
        rate_limit=1000
    )

    # Quota settings to associate with the Usage Plan for the REST API
    quota: apigateway.QuotaSettings = apigateway.QuotaSettings(
        limit=10000000,
        period=apigateway.Period.MONTH
    )


# noinspection PyArgumentList
class Http(Enum):
    """Enum class to represent an HTTP method."""
    OPTIONS = auto()
    GET = auto()
    HEAD = auto()
    PUT = auto()
    POST = auto()
    DELETE = auto()
    PATCH = auto()



"""AWS CDK Construct module."""
from __future__ import annotations

import os

from aws_cdk import (
    CfnResource,
    CfnOutput,
    aws_lambda as lambda_,
    aws_apigateway as apigateway,
)
from constructs import Construct

from .aws import SecretsManager, SSM
from .cache import APIKeyCache
from .models import Config, Http


class SecureRestApi(apigateway.RestApi):
    """Represents a Secure REST API in Amazon API Gateway.

    By default, methods in the REST API requires an ``x-api-key`` header in
    the request.

    The API key value is also populated in the CFn *stack outputs* by default,
    and shows up under the ``APIKey`` name.

    Use ``add_resource``, ``add_lambda_methods``, and ``add_methods`` to
    configure the API model.

    By default, the API will automatically be deployed and accessible from a
    public endpoint.

    Example::

        >>> state_machine = stepfunctions.StateMachine(self, "MyStateMachine",
        ...     state_machine_type=stepfunctions.StateMachineType.EXPRESS,
        ...     definition=stepfunctions.Chain.start(stepfunctions.Pass(self, "Pass"))
        >>> )

        >>> api = aws_cdk_secure_api.SecureRestApi(self, "Api",
        ...           rest_api_name="MyApi"
        ...       )
        >>> api.add_methods(apigateway.StepFunctionsIntegration.start_execution(state_machine),
        ...                 "GET", "PUT")

    """

    def __init__(self, scope: Construct, construct_id: str,
                 config: Config | None = None, **kwargs) -> None:

        super().__init__(scope, construct_id, **kwargs)

        if config is None:
            config = Config(region=self.env.region)

        elif config.region is None:
            config.region = self.env.region

        stack_name = self.stack.stack_name

        cache = APIKeyCache(self)
        api_key_value: str = cache.get_api_key()

        if api_key_value is None:
            aws_profile = (config.profile
                           or os.getenv('AWS_PROFILE')
                           or self.node.try_get_context('profile'))

            # Use SSM Parameter Store to store / retrieve the API Key, which will need
            # to be included in the 'x-api-key' header in requests to our endpoint.
            #
            # Refs:
            #   - https://docs.aws.amazon.com/cdk/latest/guide/get_ssm_value.html
            #   - https://bobbyhadz.com/blog/aws-cdk-ssm-parameters
            ssm = SSM(config.region, aws_profile)
            param_name = f'/{stack_name}/api-key'

            api_key_value: str = ssm.get_param(param_name)

            if api_key_value is None:
                # Create the SSM Parameter, as it does not exist

                # Generate the API key using `secretsmanager:GetRandomPassword`
                sm = SecretsManager(config.region, aws_profile)
                api_key_value = sm.generate_api_key(40)
                # Create the new SSM Parameter (SecureString) with the
                # auto-generated API key.
                print(
                    f'[{param_name}] Creating new SSM Parameter...'
                )
                ssm.put_param(
                    param_name, api_key_value,
                    key_id=config.key_id,
                )

            cache.save_api_key(api_key_value)

        # Create API key and add it to usage plan for the API Gateway
        #
        # Ref: https://dev.to/dvddpl/create-and-override-apikey-for-an-aws-gatewayapi-bjg
        api_key_id = f'{stack_name}-api-key'
        api_key = self.add_api_key(
            api_key_id,
            api_key_name=stack_name,
            value=api_key_value
        )
        # Create usage plan for the API
        usage_plan_name = f'{stack_name}-usage-plan'

        usage_plan = self.add_usage_plan(
            usage_plan_name,
            name=usage_plan_name,
            api_stages=[
                apigateway.UsagePlanPerApiStage(
                    api=self,
                    stage=self.deployment_stage,
                ),
            ],
            throttle=config.throttle,
            quota=config.quota,
        )

        # Add API key to the Usage Plan
        usage_plan.add_api_key(api_key)

        # Stack Outputs #

        # This is needed so the CFN Output name correctly appears for the
        # API Endpoint, as this Output is automatically generated by CDK.
        #
        # Ref: https://github.com/aws/aws-cdk/issues/15664
        if config.override_endpoint_name:
            # noinspection PyTypeChecker
            res: CfnResource = self.node.try_find_child('Endpoint')
            res.override_logical_id('APIEndpoint')

        # Ensure the API Key shows up in the Stack Outputs
        CfnOutput(
            scope, 'APIKey',
            value=api_key_value,
            # Needs to be unique per account + region
            export_name=f'x-api-key:{stack_name}'
        )

    def add_lambda_methods(self, handler: lambda_.IFunction,
                           *methods: Http | str):
        """
        Adds a lambda function the API Gateway, and associates it with one or
        more HTTP method(s).

        Example::

            >>> from aws_cdk_secure_api import SecureRestApi
            >>> from aws_cdk import (aws_apigateway as apigw, aws_lambda as lambda_)
            >>> api = SecureRestApi(self, 'api', rest_api_name='My Secure Service')
            >>> handler1 = lambda_.Function(self, 'lambda1', runtime=lambda_.Runtime.PYTHON_3_9, ...)
            >>> handler2 = lambda_.Function(self, 'lambda2', runtime=lambda_.Runtime.PYTHON_3_9, ...)
            >>> api.add_lambda_methods(handler1, 'GET')
            >>> api.add_lambda_methods(handler2, 'PUT', 'POST')

        """
        lambda_integration = apigateway.LambdaIntegration(
            handler,
            request_templates={'application/json': '{ "statusCode": "200" }'}
        )

        self.add_methods(lambda_integration, *methods)

    def add_methods(self, target: apigateway.Integration,
                    *methods: Http | str):
        """
        Adds one or more HTTP method(s) and a target (typically a
        `LambdaIntegration`) to the API root resource.

        By default, an API key is required for the new method(s).

        Example::

            >>> from aws_cdk_secure_api import SecureRestApi
            >>> from aws_cdk import (aws_apigateway as apigw, aws_lambda as lambda_)
            >>> api = SecureRestApi(self, 'api', rest_api_name='My Secure Service')
            >>> handler = lambda_.Function(self, 'my-function', runtime=lambda_.Runtime.PYTHON_3_9, ...)
            >>> integration = apigw.LambdaIntegration(
            ...   handler, request_templates={'application/json': '{ "statusCode": "200" }'}
            ... )
            >>> api.add_methods(integration, 'GET', 'POST')

        """
        if not methods:
            raise Exception('At least one HTTP method is required.')

        for method in methods:
            http_meth: str = getattr(method, 'name', method)

            self.root.add_method(
                http_meth, target,
                api_key_required=True
            )



from __future__ import annotations

import logging

from .client_cache import ClientCache


LOG = logging.getLogger('ssm')
LOG.setLevel('DEBUG')


class SSM(ClientCache):
    """
    Provides an interface to access AWS SSM Parameter Store
    """

    SERVICE_NAME = 'ssm'

    def get_param(self, name: str, with_decryption=True) -> str | None:
        """
        Get a parameter from SSM Parameter Store. Returns `None` if the
        parameter does not exist.

        Example::

        >>> from aws_cdk_secure_api.aws import SSM
        >>> ssm = SSM()
        >>> ssm.get_param('/Prod/Db/Password')

        :param name: The SSM parameter name to retrieve.
        :param with_decryption: True to decrypt a (secure string) parameter.
        :return: The decrypted parameter value, as a string.

        """
        try:
            parameter = self.client.get_parameter(
                Name=name,
                WithDecryption=with_decryption,
            )['Parameter']

            value = parameter['Value']

            if parameter['Type'] == 'StringList':
                value = value.split(',')

            return value

        except self.client.exceptions.ParameterNotFound:
            LOG.warning('Parameter `%s` does not exist in the AWS account.', name)
            return None

    def put_param(self, name: str,
                  value: str,
                  description: str | None = None,
                  type='SecureString',
                  key_id: str | None = None,
                  tier='Standard',
                  data_type='text',
                  overwrite=False) -> str:
        """
        Create/update a parameter in SSM Parameter Store.

        Example::

        >>> from aws_cdk_secure_api.aws import SSM
        >>> ssm = SSM()
        >>> ssm.put_param('/Prod/Db/Password', 'my-value')

        :param name: The fully qualified name of the parameter to store. The
          fully qualified name usually includes the complete hierarchy of the
          parameter path and name.
        :param value: specifies the parameter value. Standard parameters have
          a value limit of 4 KB, whereas Advanced parameters have a value
          limit of 8 KB.
        :param description: Specifies additional information about the
          parameter.
        :param type: The Parameter type. Must be one of
          [String, StringList, SecureString]. Defaults to `SecureString` if
          not specified.
        :param key_id: The Key Management Service (KMS) ID that you want to
          use to encrypt a parameter. Either the default KMS key automatically
          assigned to your AWS account or a custom key. Required for parameters
          that use the `SecureString` data type. If you don't specify a key ID,
          the system uses the default key associated with your AWS account.
        :param tier: Must be one of [Standard, Advanced, Intelligent-Tiering].
          Defaults to `Standard` if not specified.
        :param data_type: The data type for a String parameter. Must be one of
          [text, aws:ec2:image]. Defaults to `text` if not specified.
        :param overwrite: Used to overwrite a current parameter value. The
          default value is `False`.

        """
        kwargs = {}
        if description:
            kwargs['Description'] = description
        if key_id:
            kwargs['KeyId'] = key_id

        new_string_parameter = self.client.put_parameter(
            Name=name,
            Value=value,
            Type=type,
            Overwrite=overwrite,
            Tier=tier,
            DataType=data_type,
            **kwargs,
        )

        LOG.debug(
            'String Parameter created with version %s and Tier %s',
            new_string_parameter['Version'],
            new_string_parameter['Tier'],
        )

        return new_string_parameter



"""
AWS CDK Secure API
~~~~~~~~~~~~~~~~~~

A CDK (v2) Construct Library for Secure Rest APIs

Sample Usage:

    >>> import aws_cdk_secure_api

For full documentation and more advanced usage, please see
<https://aws-cdk-secure-api.readthedocs.io>.

:copyright: (c) 2022 by Ritvik Nag.
:license:MIT, see LICENSE for more details.
"""

__all__ = [
    'Config',
    'Http',
    'SecureRestApi',
]

import logging

from .api_construct import SecureRestApi
from .models import Config, Http

# Set up logging to ``/dev/null`` like a library is supposed to.
# http://docs.python.org/3.3/howto/logging.html#configuring-logging-for-a-library
logging.getLogger('aws_cdk_secure_api').addHandler(logging.NullHandler())



"""
AWS CDK Secure API - A CDK (v2) Construct Library for Secure REST APIs
"""

__title__ = 'aws-cdk-secure-api'
__description__ = 'A CDK (v2) Construct Library for Secure REST APIs'
__url__ = 'https://github.com/rnag/aws-cdk-secure-api'
__version__ = '0.1.1'
__author__ = 'Ritvik Nag'
__author_email__ = 'rv.kvetch@gmail.com'
__license__ = 'MIT'
__copyright__ = 'Copyright 2022 Ritvik Nag'



from .client_cache import ClientCache


class SecretsManager(ClientCache):
    """
    Provides an interface to access AWS Secrets Manager
    """

    SERVICE_NAME = 'secretsmanager'

    def generate_api_key(self, length: int = 32,
                         exclude_chars=' ^,%+~`#$&*()|[]{}:;<>?!\'/@"\\'):
        """
        Generates a random API key value of a specified length.

        :param length: The length of the API key. If you don't
          include this parameter, the default length is 32 characters.
        :param exclude_chars: The characters to exclude in the generated
          password.

        """
        res = self.client.get_random_password(
            PasswordLength=length,
            ExcludeCharacters=exclude_chars,
        )

        return res['RandomPassword']



"""The setup script."""
import pathlib

from setuptools import setup, find_packages


here = pathlib.Path(__file__).parent

package_name = 'aws_cdk_secure_api'

packages = find_packages(include=[package_name, f'{package_name}.*'])

requires = [
    'aws-cdk-lib>=2.0.0,<3.0.0',
    'constructs>=10.0.0,<11.0.0',
    'boto3>=1.0.0,<2.0.0',
]

test_requirements = [
    'pytest~=7.0.1',
    'pytest-cov~=3.0.0',
    'pytest-runner~=5.3.1']

about = {}
exec((here / package_name / '__version__.py').read_text(), about)

readme = (here / 'README.rst').read_text()
history = (here / 'HISTORY.rst').read_text()

setup(
    name=about['__title__'],
    version=about['__version__'],
    description=about['__description__'],
    long_description=readme + '\n\n' + history,
    long_description_content_type='text/x-rst',
    author=about['__author__'],
    author_email=about['__author_email__'],
    url=about['__url__'],
    packages=packages,
    include_package_data=True,
    install_requires=requires,
    project_urls={
        'Documentation': 'https://aws-cdk-secure-api.readthedocs.io',
        'Source': 'https://github.com/rnag/aws-cdk-secure-api',
    },
    license=about['__license__'],
    # TODO add more relevant keywords as needed
    keywords=['aws-cdk', 'secure', 'rest-api', 'api-gateway', 'api key', 'cdk'],
    classifiers=[
        # Ref: https://pypi.org/classifiers/
        # 'Development Status :: 5 - Production/Stable',
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python'
    ],
    test_suite='tests',
    tests_require=test_requirements,
    zip_safe=False
)



from collections import defaultdict
from dataclasses import dataclass, InitVar
from typing import ClassVar, DefaultDict, Optional

from boto3 import Session, client
from botocore.client import BaseClient
from botocore.config import Config
from botocore.exceptions import UnknownServiceError


@dataclass
class ClientCache:

    # boto3 service name
    SERVICE_NAME: ClassVar[str] = None

    # sub-classes can specify that client objects should be thread-safe
    THREAD_SAFE: ClassVar[bool] = False

    _clients: ClassVar[DefaultDict] = defaultdict(dict)

    # The specified region that is bound to a client. Default to 'us-east-1'.
    region_name: str = 'us-east-1'

    # Optional AWS profile name.
    profile_name: Optional[str] = None

    # When enabled, automatically initializes the client for the region; this
    # is useful when service requests will be made in multiple threads and it
    # is desirable to reuse the same client between threads.
    init_client: InitVar[bool] = False

    # Maximum pool connections for multi-thread usage
    # Ref: https://stackoverflow.com/a/68760777/10237506
    max_pool_connections: Optional[int] = None

    def __post_init__(self, init_client: bool):
        self.region_name = self.region_name.lower()

        if init_client:
            _ = self.client

    @property
    def client(self):
        return self._get_client(self.region_name)

    @client.setter
    def client(self, value):
        raise Exception('Member read-only')

    def _get_client(self, region_name):
        """
        Internal method to return a low-level SecretManager client for a given region name
        """
        if region_name not in self._clients[self.SERVICE_NAME]:
            self._clients[self.SERVICE_NAME][region_name] = self._create_client()

        return self._clients[self.SERVICE_NAME][region_name]

    def _create_client(self) -> BaseClient:
        if not self.SERVICE_NAME:
            raise ValueError(
                'Sub-classes must provide a value for "SERVICE_NAME"')

        try:
            if self.THREAD_SAFE or self.profile_name:
                client_func = Session(profile_name=self.profile_name).client
            else:
                client_func = client

            client_kwargs = {}
            if self.max_pool_connections:
                client_kwargs['config'] = Config(
                    max_pool_connections=self.max_pool_connections)

            return client_func(self.SERVICE_NAME, self.region_name,
                               **client_kwargs)

        except UnknownServiceError:
            raise NotImplementedError(
                'Sub-classes must override this method')



"""Tests for `aws_cdk_secure_api` package."""

import pytest


from aws_cdk_secure_api import SecureRestApi


@pytest.fixture
def response():
    """Sample pytest fixture.

    See more at: http://doc.pytest.org/en/latest/fixture.html
    """
    # import requests
    # return requests.get('https://github.com/audreyr/cookiecutter-pypackage')


def test_create_api(response):
    """Sample pytest test function with the pytest fixture as an argument."""
    # TODO
    SecureRestApi
